@@ -33,18 +33,28 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <unistd.h>
+#include <netdb.h>
+#include <net/if.h>
+#include <arpa/inet.h>
 
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
 #include "vino-util.h"
+#include "vino-mdns.h"
 #ifdef VINO_ENABLE_HTTP_SERVER
 #include "vino-http.h"
 #endif
 
-#define VINO_DBUS_LISTENER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o),                     \
-                                                                        VINO_TYPE_DBUS_LISTENER, \
-                                                                        VinoDBusListenerPrivate))
+#ifdef HAVE_IFADDRS_H
+#include <ifaddrs.h>
+#else
+#include "libvncserver/ifaddr/ifaddrs.h"
+#endif
+
+#define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
+#define VINO_DBUS_BUS_NAME  "org.gnome.Vino"
 
 G_DEFINE_TYPE (VinoDBusListener, vino_dbus_listener, G_TYPE_OBJECT)
 
@@ -62,6 +72,89 @@ enum
 static void vino_dbus_listener_set_server (VinoDBusListener *listener,
                                            VinoServer       *server);
 
+static char *
+get_local_hostname (VinoDBusListener *listener)
+{
+  char                *retval, buf[INET6_ADDRSTRLEN];
+  struct ifaddrs      *myaddrs, *ifa; 
+  void                *sin;
+  const char          *server_iface;
+  GHashTable          *ipv4, *ipv6;
+  GHashTableIter      iter;
+  gpointer            key, value;
+
+  retval = NULL;
+  server_iface = vino_server_get_network_interface (listener->priv->server);
+  ipv4 = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
+  ipv6 = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
+
+  getifaddrs (&myaddrs);
+  for (ifa = myaddrs; ifa != NULL; ifa = ifa->ifa_next)
+    {
+      if (ifa->ifa_addr == NULL || ifa->ifa_name == NULL || (ifa->ifa_flags & IFF_UP) == 0)
+	continue;
+
+      switch (ifa->ifa_addr->sa_family)
+	{
+	  case AF_INET:
+	    sin = &((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
+	    inet_ntop (AF_INET, sin, buf, INET6_ADDRSTRLEN);
+	    g_hash_table_insert (ipv4,
+				 ifa->ifa_name,
+				 g_strdup (buf));
+	    break;
+
+	  case AF_INET6:
+	    sin = &((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr;
+	    inet_ntop (AF_INET6, sin, buf, INET6_ADDRSTRLEN);
+	    g_hash_table_insert (ipv6,
+				 ifa->ifa_name,
+				 g_strdup (buf));
+	    break;
+	  default: continue;
+	}
+    }
+
+  if (server_iface && server_iface[0] != '\0')
+    {
+      if ((retval = g_strdup (g_hash_table_lookup (ipv4, server_iface))))
+	goto the_end;
+      if ((retval = g_strdup (g_hash_table_lookup (ipv6, server_iface))))
+	goto the_end;
+    }
+
+  g_hash_table_iter_init (&iter, ipv4);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      if (strcmp (key, "lo") == 0)
+	continue;
+      retval = g_strdup (value);
+      goto the_end;
+    }
+
+  g_hash_table_iter_init (&iter, ipv6);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      if (strcmp (key, "lo") == 0)
+	continue;
+      retval = g_strdup (value);
+      goto the_end;
+    }
+
+  if ((retval = g_strdup (g_hash_table_lookup (ipv4, "lo"))))
+    goto the_end;
+  if ((retval = g_strdup (g_hash_table_lookup (ipv6, "lo"))))
+    goto the_end;
+
+  the_end:
+  freeifaddrs (myaddrs); 
+  g_hash_table_destroy (ipv4);
+  g_hash_table_destroy (ipv6);
+
+  return retval;
+}
+
+
 static void
 vino_dbus_listener_set_property (GObject       *object,
                                  guint          prop_id,
@@ -103,7 +196,7 @@ vino_dbus_listener_get_property (GObject    *object,
 static void
 vino_dbus_listener_init (VinoDBusListener *listener)
 {
-  listener->priv = VINO_DBUS_LISTENER_GET_PRIVATE (listener);
+  listener->priv = G_TYPE_INSTANCE_GET_PRIVATE (listener, VINO_TYPE_DBUS_LISTENER, VinoDBusListenerPrivate);
 }
 
 static void
@@ -149,16 +242,16 @@ static const char * introspect_xml =
   "    </method>\n"
   "  </interface>\n"
   "  <interface name=\"org.gnome.VinoScreen\">\n"
-  "    <method name=\"GetServerPort\">\n"
-  "      <arg name=\"port\" direction=\"out\" type=\"u\"/>\n"
+  "    <method name=\"GetInternalData\">\n"
+  "      <arg name=\"hostname\" direction=\"out\" type=\"s\"/>\n"
+  "      <arg name=\"avahi_hostname\" direction=\"out\" type=\"s\"/>\n"
+  "      <arg name=\"port\" direction=\"out\" type=\"d\"/>\n"
   "    </method>\n"
-  "    <signal name=\"ServerPortChanged\">\n"
+  "    <method name=\"GetExternalPort\">\n"
+  "      <arg name=\"port\" direction=\"out\" type=\"d\"/>\n"
+  "    </method>\n"
+  "    <signal name=\"ServerInfoChanged\">\n"
   "    </signal>\n"
-#ifdef VINO_ENABLE_HTTP_SERVER
-  "    <method name=\"GetHttpServerPort\">\n"
-  "      <arg name=\"port\" direction=\"out\" type=\"u\"/>\n"
-  "    </method>\n"
-#endif
   "  </interface>\n"
   "</node>\n";
 
@@ -196,13 +289,13 @@ vino_dbus_listener_handle_get_server_port (VinoDBusListener *listener,
                                            DBusConnection   *connection,
                                            DBusMessage      *message)
 {
-  DBusMessage  *reply;
-  dbus_int32_t  port;
+  DBusMessage *reply;
+  gint        port;
 
   if (!(reply = dbus_message_new_method_return (message)))
     goto oom;
 
-  port = vino_server_get_port (listener->priv->server);
+  port = vino_server_get_external_port (listener->priv->server);
 
   if (!dbus_message_append_args (reply, DBUS_TYPE_INT32, &port, DBUS_TYPE_INVALID))
     goto oom;
@@ -221,21 +314,32 @@ vino_dbus_listener_handle_get_server_port (VinoDBusListener *listener,
   return DBUS_HANDLER_RESULT_NEED_MEMORY;
 }
 
-#ifdef VINO_ENABLE_HTTP_SERVER
 static DBusHandlerResult
-vino_dbus_listener_handle_get_http_server_port (VinoDBusListener *listener,
-                                                DBusConnection   *connection,
-                                                DBusMessage      *message)
+vino_dbus_listener_handle_get_internal_data (VinoDBusListener *listener,
+                                             DBusConnection   *connection,
+                                             DBusMessage      *message)
 {
-  DBusMessage  *reply;
-  dbus_int32_t  port;
+  DBusMessage *reply;
+  gint        port;
+  char        *host = NULL;
+  const char  *avahi_host;
 
   if (!(reply = dbus_message_new_method_return (message)))
     goto oom;
 
-  port = vino_get_http_server_port();
+#ifdef VINO_ENABLE_HTTP_SERVER
+  port = vino_get_http_server_port (listener->priv->server);
+#else
+  port = vino_server_get_port (listener->priv->server);
+#endif
 
-  if (!dbus_message_append_args (reply, DBUS_TYPE_INT32, &port, DBUS_TYPE_INVALID))
+  host = get_local_hostname (listener);
+  avahi_host = vino_mdns_get_hostname ();
+  if (!dbus_message_append_args (reply,
+                                 DBUS_TYPE_STRING, &host,
+                                 DBUS_TYPE_STRING, &avahi_host,
+                                 DBUS_TYPE_INT32, &port,
+                                 DBUS_TYPE_INVALID))
     goto oom;
 
   if (!dbus_connection_send (connection, reply, NULL))
@@ -243,23 +347,22 @@ vino_dbus_listener_handle_get_http_server_port (VinoDBusListener *listener,
     
   dbus_message_unref (reply);
 
-  dprintf (DBUS, "Successfully handled '%s' message: port = %d\n", dbus_message_get_member (message), port);
+  dprintf (DBUS, "Successfully handled '%s' message\n", dbus_message_get_member (message));
 
   return DBUS_HANDLER_RESULT_HANDLED;
 
  oom:
+
+  g_free (host);
   g_error (_("Out of memory handling '%s' message"), dbus_message_get_member (message));
   return DBUS_HANDLER_RESULT_NEED_MEMORY;
 }
-#endif /* VINO_ENABLE_HTTP_SERVER */
 
 static DBusHandlerResult
 vino_dbus_listener_message_handler (DBusConnection *connection,
                                     DBusMessage    *message,
                                     void           *user_data)
 {
-#define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
-
   VinoDBusListener *listener = VINO_DBUS_LISTENER (user_data);
 
   dprintf (DBUS, "D-Bus message: obj_path = '%s' interface = '%s' method = '%s' destination = '%s'\n",
@@ -270,22 +373,20 @@ vino_dbus_listener_message_handler (DBusConnection *connection,
 
   if (dbus_message_is_method_call (message,
                                    VINO_DBUS_INTERFACE,
-                                   "GetServerPort"))
+                                   "GetInternalData"))
+    {
+      return vino_dbus_listener_handle_get_internal_data (listener,
+                                                          connection,
+                                                          message);
+    }
+  if (dbus_message_is_method_call (message,
+                                   VINO_DBUS_INTERFACE,
+                                   "GetExternalPort"))
     {
       return vino_dbus_listener_handle_get_server_port (listener,
                                                         connection,
                                                         message);
     }
-#ifdef VINO_ENABLE_HTTP_SERVER
-  else if (dbus_message_is_method_call (message,
-                                   VINO_DBUS_INTERFACE,
-                                   "GetHttpServerPort"))
-    {
-      return vino_dbus_listener_handle_get_http_server_port (listener,
-                                                             connection,
-                                                             message);
-    }
-#endif
   else if (dbus_message_is_method_call (message,
                                         "org.freedesktop.DBus.Introspectable",
                                         "Introspect"))
@@ -298,24 +399,22 @@ vino_dbus_listener_message_handler (DBusConnection *connection,
     {
       return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
     }
-
-#undef VINO_DBUS_INTERFACE
 }
 
 static void
-vino_dbus_listener_port_changed (VinoServer *server, VinoDBusListener *listener)
+vino_dbus_listener_info_changed (VinoServer *server, VinoDBusListener *listener)
 {
   DBusMessage *message;
   gchar *obj_path;
 
-  dprintf (DBUS, "Emitting ServerPortChanged signal\n");
+  dprintf (DBUS, "Emitting ServerInfoChanged signal\n");
 
   obj_path = g_strdup_printf ("/org/gnome/vino/screens/%d",
                               gdk_screen_get_number (vino_server_get_screen (server)));
 
   message = dbus_message_new_signal (obj_path,
-                                     "org.gnome.VinoScreen",
-                                     "ServerPortChanged");
+                                     VINO_DBUS_INTERFACE,
+                                     "ServerInfoChanged");
   g_free (obj_path);
 
   if (!message)
@@ -369,12 +468,8 @@ vino_dbus_listener_set_server (VinoDBusListener *listener,
   dprintf (DBUS, "Object registered at path '%s'\n", obj_path);
 
   g_signal_connect (server, "notify::alternative-port",
-		    G_CALLBACK (vino_dbus_listener_port_changed),
+		    G_CALLBACK (vino_dbus_listener_info_changed),
 		    listener);
-  g_signal_connect (server, "notify::use-alternative-port",
-		    G_CALLBACK (vino_dbus_listener_port_changed),
-		    listener);
-
   g_free (obj_path);
 }
 
@@ -428,7 +523,6 @@ vino_dbus_unref_connection (void)
 gboolean
 vino_dbus_request_name (void)
 {
-#define VINO_DBUS_BUS_NAME "org.gnome.Vino"
 
   DBusConnection *connection;
   DBusError       error;
@@ -459,6 +553,4 @@ vino_dbus_request_name (void)
 
   dprintf (DBUS, "Successfully acquired D-Bus name '%s'\n", VINO_DBUS_BUS_NAME);
   return TRUE;
-
-#undef VINO_DBUS_BUS_NAME
 }
